(*
This file is part of Ceve.

Ceve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Ceve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Ceve; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(**
Main module
@author Jaap Boender
*)

open Metadata
open Options

(**
Read the input (whatever type it is) into a list of package metadata
@param files The list of files to be read
@return a list of package metadata
*)
let read_input (files: string list): pkg_metadata list =
begin
	match !input_type with
	|	Debian -> List.map Deb_reader.read_deb_file files
	| DebianCache -> List.flatten (List.map Deb_reader.read_pool_file files)
	| RPM -> List.map (Rpm_reader.read_rpm_file "") files
	| HDList -> List.flatten (List.map (Rpm_reader.read_hdlist "") files)
	| EGraphIn -> List.flatten (List.map Egraph_reader.read_egraph_file files)
end;;

(**
Output the metadata in the right format
@param metadata The metadata
*)
let do_output (metadata: pkg_metadata list): unit =
begin
	match !output_type with
	| PrettyPrint -> Pretty_print.pretty_print metadata
	| EGraphOut -> Egraph_writer.output_egraph metadata
	(* | RPMFind -> Rpmfind_writer.output_rpmfind metadata *)
	| Dose -> Dose_writer.output_dose metadata
	| Oz -> Oz_writer.output_oz metadata
	| Graphviz -> Graphviz_writer.output_graphviz metadata
	| TartCNF -> Cnf_writer.output_tart_cnf metadata
end;;

(** main function *)
let _ =
let pkgs = ref [] in
begin
	if !verbose then prerr_endline "[Main] Starting ceve.";
	parse_options;
	if !verbose then prerr_endline "[Main] Options parsed.";
	pkgs := read_input (!input_files);
	if !verbose then prerr_endline "[Main] Input read.";
	if !version_type = RPMVersions then
		pkgs := Dependencies.add_file_provides !pkgs;
	if !resolve_dependencies then
		pkgs := Dependencies.resolve_dependencies !pkgs;
	begin
		match !cone_package with
		| Some p -> pkgs := (Dependencies.extract_cone !pkgs p)
		| _ -> ()
	end;
	do_output !pkgs;
	if !verbose then prerr_endline "[Main] Output written."
end;;
