(*
This file is part of Ceve.

Ceve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Ceve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Ceve; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

(**
Module to parse Debian modules
@author Jaap Boender
*)

open Big_int
open Metadata

let buffer = ref '\x00'

(**
Read a character from a .deb file (buffer if necessary)
@param channel Channel to read from
@param input_char Function that reads a character from the input channel
@return the character read
*)
let read_deb_char (channel: 'a) (input_char: 'a -> char): char =
begin
	if !buffer = '\x00' then
		buffer := input_char channel;
	let result = !buffer in
		buffer := input_char channel; result
end;;

(**
Parse a debian name and version specification
Example: kde (>= 3.5)
@param spec The specification
@return a pkg_dependency containing the specification
*)
let parse_deb_name_version (spec: string): pkg_dependency =
begin
	try
	Scanf.sscanf (Tools.truncate spec) "%[^(](%[^)])" (fun t v -> {
		target = (Tools.truncate t);
		version_spec = 
			(if (String.sub v 0 1) = "=" then
				Equal (Tools.truncate (String.sub v 1 ((String.length v) - 1)))
			else if (String.sub v 0 2) = "<<" then
				Less (Tools.truncate (String.sub v 2 ((String.length v) - 2)))
			else if (String.sub v 0 2) = "<=" then
				LessEqual (Tools.truncate (String.sub v 2 ((String.length v) - 2)))
			else if (String.sub v 0 2) = ">=" then
				GreaterEqual (Tools.truncate (String.sub v 2 ((String.length v) - 2)))
			else if (String.sub v 0 2) = ">>" then
				Greater (Tools.truncate (String.sub v 2 ((String.length v) - 2)))
			else if (String.sub v 0 1) = "<" then
				Less (Tools.truncate (String.sub v 1 ((String.length v) - 1)))
			else if (String.sub v 0 1) = ">" then
				Greater (Tools.truncate (String.sub v 1 ((String.length v) - 1)))
			else
				AnyVersion
			)
	})
	with End_of_file ->
		{target = Tools.truncate spec; version_spec = AnyVersion }
end;;

(**
Parse a .deb dependency line (Depends: ...)
@param line The dependency line
@return a pkg_dep_expression containing the line
*)
let parse_deb_dependencies (line: string): pkg_dep_expression =
let disjunctions = Str.split (Str.regexp_string ",") line in
begin
	List.map (fun disj ->
		List.map parse_deb_name_version (List.map Tools.truncate (Str.split (Str.regexp_string "|") disj))
	) disjunctions
end;;

(**
Read a field from a .deb file
@param channel Channel to read from
@param input_char Function to read a character from that channel
*)
let read_deb_field (channel: 'a) (input_char: 'a -> char) (old_metadata: pkg_metadata): pkg_metadata =
let my_char = ref ' '
and my_line = ref ""
and is_multiline = ref true
and my_name = ref ""
and my_contents = ref "" in
begin
	while !is_multiline
	do
		my_char := read_deb_char channel input_char;
		while !my_char <> '\n'
		do
			my_line := !my_line ^ (String.make 1 !my_char);
			my_char := read_deb_char channel input_char;
		done;
		is_multiline := (!buffer = ' ');
		if (!is_multiline) then my_line := !my_line ^ "\n"
	done;		
	if !my_line = "" then raise End_of_file;
	let colon_index = String.index  !my_line ':' in
	my_name := String.lowercase (String.sub !my_line 0 colon_index);
	my_contents := Tools.truncate (String.sub !my_line (colon_index + 1) ((String.length !my_line) - colon_index - 1));
	if !my_name = "architecture" then
	let architecture = String.lowercase !my_contents in
	begin
		{old_metadata with architecture = if architecture = "all" then
			AllArchitectures
		else if architecture = "any" then
			AnyArchitecture
		else if architecture = "source" then
			Source
		else
			Specific architecture}
	end
	else if !my_name = "bugs" then
		{old_metadata with bugs = !my_contents}
	else if !my_name = "build-essential" then
		{old_metadata with build_essential = (String.lowercase !my_contents = "yes")}
	else if !my_name = "conflicts" then
		{old_metadata with conflicts = List.map parse_deb_name_version (Str.split (Str.regexp_string ",") !my_contents)}
	else if !my_name = "depends" then
		{old_metadata with depends = parse_deb_dependencies !my_contents} 	
	else if !my_name = "description" then
	try
		let newline_index = String.index !my_contents '\n' in
			{old_metadata with
				summary = String.sub !my_contents 0 newline_index;
				description = String.sub !my_contents (newline_index + 1) ((String.length !my_contents) - newline_index - 1)}
	with Not_found -> {old_metadata with summary = !my_contents}
	else if !my_name = "enhances" then
		{old_metadata with enhances = parse_deb_dependencies !my_contents}
	else if !my_name = "essential" then
		{old_metadata with essential = (String.lowercase !my_contents = "yes")}
	else if !my_name = "filename" then
		{old_metadata with filename = !my_contents}
	else if !my_name = "installed-size" then
		{old_metadata with installed_size = Int32.of_string !my_contents}
	else if !my_name = "maintainer" then
		{old_metadata with maintainer = !my_contents}
	else if !my_name = "md5sum" then
		{old_metadata with md5sum = !my_contents}
	else if !my_name = "package" then
	begin
		if (!Options.verbose) then
			prerr_endline ("[Deb] Parsing package " ^ !my_contents);
		{old_metadata with name = !my_contents}
	end
	else if !my_name = "origin" then
		{old_metadata with origin = !my_contents}
	else if !my_name = "pre-depends" then
		{old_metadata with pre_depends = parse_deb_dependencies !my_contents}
	else if !my_name = "provides" then
		{old_metadata with provides = List.map (fun x -> {target = (Tools.truncate x); version_spec = AnyVersion}) (Str.split (Str.regexp_string ",") !my_contents)}
	else if !my_name = "priority" then
	let priority = String.lowercase !my_contents in
	begin
		{old_metadata with priority = if priority = "required" then
			Required
		else if priority = "important" then
			Important
		else if priority = "standard" then
			Standard
		else if priority = "optional" then
			Optional
		else if priority = "extra" then
			Extra
		else
			raise (Failure ("Unknown priority " ^ !my_contents))}
	end
	else if !my_name = "python-version" then
		{old_metadata with python_version = !my_contents}
	else if !my_name = "recommends" then
		{old_metadata with recommends = parse_deb_dependencies !my_contents}
	else if !my_name = "replaces" then
		{old_metadata with replaces = List.map parse_deb_name_version (Str.split (Str.regexp_string ",") !my_contents)}
	else if !my_name = "section" then
		{old_metadata with section = !my_contents}
	else if !my_name = "sha1" then
		{old_metadata with sha1sum = !my_contents}
	else if !my_name = "sha256" then
		{old_metadata with sha256sum = !my_contents}
	else if !my_name = "size" then
		{old_metadata with size = Int32.of_string !my_contents}
	else if !my_name = "source" then
	let src_name = ref ""
	and src_version = ref "" in
		try
			Scanf.sscanf !my_contents "%[^ \t]%_[ \t](%[^)])" (fun n v -> src_name := n; src_version := v);
			{old_metadata with source_package = !src_name; source_version = !src_version}
		with
			End_of_file -> {old_metadata with source_package = !my_contents}
	else if !my_name = "suggests" then
		{old_metadata with suggests = parse_deb_dependencies !my_contents}
	else if !my_name = "tag" then
		{old_metadata with tags = List.map Tools.truncate (Str.split (Str.regexp_string ",") !my_contents)}
	else if !my_name = "task" then
		{old_metadata with tasks = List.map Tools.truncate (Str.split (Str.regexp_string ",") !my_contents)}
	else if !my_name = "url" then
		{old_metadata with url = !my_contents}
	else if !my_name = "version" then
		{old_metadata with version = !my_contents}
	else
	begin
		prerr_endline ("Warning: unknown field " ^ !my_name ^ "!");
		{old_metadata with unknown = (!my_name, !my_contents)::old_metadata.unknown}
	end
end;;

(** Read MD5 sums from a .deb file
@param channel Channel to input from
@param input_char Function to read a character from that channel
@param size The size of the MD5 file in the .deb file
@return a list of tuples (name and MD5 sum)
*)
let read_deb_md5sums (channel: 'a) (input_char: 'a -> char) (size: big_int): (string * string) list =
let files = ref []
and my_char = ref (input_char channel) 
and my_line = ref "" 
and chars_read = ref unit_big_int in
begin
	try
		while (le_big_int !chars_read size)
		do
			while (!my_char <> '\n')
			do
				my_line := !my_line ^ (String.make 1 !my_char);
				my_char := input_char channel;
				chars_read := succ_big_int !chars_read;
			done;
			Scanf.sscanf !my_line "%s %s" (fun s f ->
				files := (f, s)::!files);
			my_line := "";
			my_char := input_char channel;
			chars_read := succ_big_int !chars_read
		done;
		(* read up to 512 extra characters *)
		let rest = 512 - (int_of_big_int (mod_big_int size (big_int_of_int 512))) in
			if rest <> 0 then for i = 2 to rest do ignore (input_char channel) done;
		!files
	with
		End_of_file -> !files
end;;

(**
Read a .deb file
@param filename File name
@return package metadata
*)
let read_deb_file (filename: string): pkg_metadata =
let ar_channel = Pirate.open_ar_file filename 
and package = ref empty_metadata in 
begin
	prerr_endline ("[Deb] Reading Debian package " ^ filename ^ "...");
	ignore (Pirate.ar_seek_to_file ar_channel "control.tar.gz");
	let gz_channel = Gzip.open_in_chan ar_channel in
	let sums_size = Pirate.tar_seek_to_file gz_channel Gzip.input Gzip.input_char "./md5sums" in
	package := {!package with files = read_deb_md5sums gz_channel Gzip.input_char sums_size}; 
	ignore (Pirate.tar_seek_to_file gz_channel Gzip.input Gzip.input_char "./control");
	try
		while true
		do
			package := read_deb_field gz_channel Gzip.input_char !package
		done;
		!package
	with
		End_of_file -> !package
end;;

(**
Read a pool (cache) file
@param filename File name
@return a list package metadata
*)
let read_pool_file (filename: string): pkg_metadata list =
let channel = open_in filename in
let package = ref empty_metadata 
and packages = ref [] in
begin
	prerr_endline ("[Deb] Reading Debian cache/pool file " ^ filename ^ "...");
	try
		package := read_deb_field channel Pervasives.input_char !package;
		while !package <> empty_metadata
		do
			try
				package := read_deb_field channel Pervasives.input_char !package
			with
				End_of_file ->
					packages := (!package::!packages);
					package := read_deb_field channel Pervasives.input_char empty_metadata
		done;
		!packages
	with End_of_file -> !packages
end;;
