(*
This file is part of Ceve.

Ceve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Ceve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Ceve; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

open Options
open Metadata

type vp_hashtable = (string, pkg_version_spec * pkg_dependency) Hashtbl.t
type pkg_hashtable = (string, string * pkg_metadata) Hashtbl.t

let version_of (spec: pkg_version_spec): string =
begin
	match spec with
	  AnyVersion -> ""
	| Less x | LessEqual x | Equal x | GreaterEqual x | Greater x -> x
end;;

let split_version (v: string): string * string * string =
let split_epoch (v: string): string * string =
begin
	try
		let colon_index = String.index v ':' in
			(String.sub v 0 colon_index, String.sub v (colon_index+1) ((String.length v)-colon_index-1)) 
	with
		Not_found -> ("0", v)
end
and split_release (v: string): string * string =
begin
	try
		let hyphen_index = String.rindex v '-' in
			(String.sub v 0 hyphen_index, String.sub v (hyphen_index+1) ((String.length v)-hyphen_index-1))
	with
		Not_found -> (v, "")
end in
begin
	let (epoch, rest) = split_epoch v in
	let (upstream, release) = split_release rest in
		(epoch, upstream, release)
end;;

let debian_compare_versions (v1: string) (v2: string): int =
let rec modified_compare (s1: string) (s2: string): int =
let letters = "abcdefghijklmnopqrtstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ" in
begin
	(* prerr_endline ("modified_compare [" ^ s1 ^ "] [" ^ s2 ^ "]"); *)
	if s1 = s2 then 0
	else if s1 = "" then
	begin
		if s2.[0] = '~' then 1
		else -1
	end
	else if s1.[0] = '~' then
	begin
		if s2 = "" then -1 (* [A] *)
		else if s2.[0] = '~' then (* [B] *) 
			modified_compare (String.sub s1 1 ((String.length s1)-1)) (String.sub s2 1 ((String.length s2)-1))
		else -1
	end
	else
	begin
		if s2 = "" then 1
		(* s1.[0] cannot be ~ we'd have found that already at [A] *)
		else if s2.[0] = '~' then 1
		(* s1.[0] cannot be ~, we'd have found that already at [B] *)
		else
		begin
			if s1.[0] = s2.[0] then
				modified_compare (String.sub s1 1 ((String.length s1)-1)) (String.sub s2 1 ((String.length s2)-1))
			else
			begin
				if String.contains letters s1.[0] then
				begin
					if String.contains letters s2.[0] then compare s1.[0] s2.[0]
					else -1
				end
				else
				begin
					if String.contains letters s2.[0] then 1
					else compare s1.[0] s2.[0]
				end
			end
		end
	end
end in
let rec compare_interior (i1: string) (i2: string): int =
begin
	(* prerr_endline ("compare_interior [" ^ i1 ^ "] [" ^ i2 ^ "]"); *)
	if i1 = i2 then 0
	else
	let (nd1, d1, r1) = Scanf.sscanf i1 "%[^0-9]%[0-9]%s%!" (fun x y z -> (x, (if y = "" then Int64.zero else Int64.of_string y), z))
	and (nd2, d2, r2) = Scanf.sscanf i2 "%[^0-9]%[0-9]%s%!" (fun x y z -> (x, (if y = "" then Int64.zero else Int64.of_string y), z)) in
	let nondigits = modified_compare nd1 nd2 in
		if nondigits = 0 then
		begin
			let digits = Int64.compare d1 d2 in
			if digits = 0 then
				compare_interior r1 r2
			else
				digits
		end
		else
			nondigits
end in 
let (e1, u1, r1) = split_version v1
and (e2, u2, r2) = split_version v2 in
begin
	(* prerr_endline ("v1: [" ^ e1 ^ "] [" ^ u1 ^ "] [" ^ r1 ^ "]");
	prerr_endline ("v2: [" ^ e2 ^ "] [" ^ u2 ^ "] [" ^ r2 ^ "]"); *)
	let epochs = compare (int_of_string e1) (int_of_string e2) in
	if epochs = 0 then
	let upstreams = compare_interior u1 u2 in
	begin
		if upstreams = 0 then compare_interior r1 r2
		else upstreams
	end
	else
	  epochs
end;;
let rpm_compare_versions (v1: string) (v2: string): int =
let rec get_chunk (s: string): string * string =
let chunk = ref "" 
and rest = ref "" in
begin
	Scanf.sscanf s "%_[^0-9a-zA-Z]%[0-9]%s%!" (fun x y -> chunk := x; rest := y);
	if !chunk = "" then (* no numeric chunk *)
		Scanf.sscanf s "%_[^0-9a-zA-Z]%[a-zA-Z]%s%!" (fun x y -> chunk := x; rest := y);
	(!chunk, !rest)
end in
let compare_ztrange (s1: string) (s2: string): int =
begin
	if String.contains "0123456789" s1.[0] then (* s1 is numeric *)
		1 (* s2 cannot be numeric *)
	else (* s1 is NOT numeric *)
	begin
		if String.contains "0123456789" s2.[0] then (* s2 is numeric *)
			-1
		else
			compare s1 s2
	end
end in
let rec compare_chunky (s1: string) (s2: string): int =
let (c1, r1) = get_chunk s1
and (c2, r2) = get_chunk s2 in
begin
	if c1 = "" && r1 = "" && c2 = "" && r2 = "" then 0
	else if c1 = "" && c2 <> "" then -1
	else if c1 <> "" && c2 = "" then 1
	else if c1 = c2 then compare_chunky r1 r2
	else
		let res = try
			compare (int_of_string c1) (int_of_string c2)
		with Failure _ -> compare_ztrange c1 c2 in
		if res = 0 then compare_chunky r1 r2
		else res
end in
let (e1, u1, r1) = split_version v1
and (e2, u2, r2) = split_version v2 in
begin
	let epochs = compare (int_of_string e1) (int_of_string e2) in
	if epochs = 0 then
	begin
		let upstreams = compare_chunky u1 u2 in
		if upstreams = 0 then
		begin
			if r1 = "" then 0
			else if r2 = "" then 0
			else compare_chunky r1 r2
		end
		else upstreams
	end
	else
	  epochs
end;;

let compare_versions = match !version_type with
| RPMVersions -> rpm_compare_versions
| _ -> debian_compare_versions;;

(**
Find out whether a package (with a certain version) can provide a dependency.
For example: [can_provide (Equal "6.3") (Less "6.4")] returns true, since
version 6.3 satisfies the condition "less than 6.4".
*)
let can_provide (provider: pkg_version_spec) (dependency: pkg_version_spec): bool =
begin
	match dependency with
	| AnyVersion -> true (* any version can be satisfied by anything *) 
	| Equal d -> (match provider with
		| AnyVersion -> true
		| Equal p -> compare_versions d p = 0
		| Less p -> compare_versions d p < 0
		| LessEqual p -> compare_versions d p <= 0
		| Greater p -> compare_versions d p > 0
		| GreaterEqual p -> compare_versions d p >= 0)
	| Less d -> (match provider with
		| AnyVersion -> true
		| Equal p -> compare_versions p d < 0
		| Less p | LessEqual p -> true
		| Greater p | GreaterEqual p -> compare_versions p d < 0)
	| LessEqual d -> (match provider with
		| AnyVersion -> true
		| Equal p -> compare_versions p d <= 0
		| Less p | LessEqual p -> true
		| Greater p -> compare_versions p d < 0
		| GreaterEqual p -> compare_versions p d <= 0)
	| Greater d -> (match provider with
		| AnyVersion -> true
		| Equal p -> compare_versions p d > 0
		| Less p | LessEqual p -> compare_versions p d > 0
		| Greater p | GreaterEqual p -> true)
	| GreaterEqual d -> (match provider with
		| AnyVersion -> true
		| Equal p -> compare_versions p d >= 0
		| Less p -> compare_versions p d > 0
		| LessEqual p -> compare_versions p d >= 0
		| Greater p | GreaterEqual p -> true)
end;;

let create_vp_hashtable (pkgs: pkg_metadata list): vp_hashtable =
let ht = Hashtbl.create (List.length pkgs) in
begin
	if !verbose then prerr_endline "[Deps] Creating virtual package hashtable...";
	List.iter (fun pkg ->
		List.iter (fun provide ->
			Hashtbl.add ht provide.target (provide.version_spec, { target = pkg.name; version_spec = Equal pkg.version })
		) pkg.provides;
		(* And for Debian, packages provide themselves *)
		if !Options.version_type = Options.DebianVersions then
			Hashtbl.add ht pkg.name (Equal pkg.version, { target = pkg.name; version_spec = Equal pkg.version })
	) pkgs;
	ht
end;;

let create_pkgs_hashtable (pkgs: pkg_metadata list): pkg_hashtable =
let ht = Hashtbl.create (List.length pkgs) in
begin
	if !verbose then prerr_endline "[Deps] Creating package hashtable...";
	List.iter (fun pkg -> Hashtbl.add ht pkg.name (pkg.version, pkg)) pkgs;
	ht
end;;

let get_dependencies (pkg: pkg_metadata): pkg_dependency list =
let result = ref [] in
begin
	List.iter (fun x ->
		match x with
			Conflicts -> result := !result @ pkg.conflicts
		| Depends -> result := !result @ (List.flatten pkg.depends)
		| Enhances -> result := !result @ (List.flatten pkg.enhances)
		| PreDepends -> result := !result @ (List.flatten pkg.pre_depends)
		| Recommends -> result := !result @ (List.flatten pkg.recommends)
		| Replaces -> result := !result @ pkg.replaces
		| Suggests -> result := !result @ (List.flatten pkg.suggests)
	) !Options.cone_types;
	!result
end;;

let replace_virtual_dependencies (name, version) (dep_type: string) (pht: pkg_hashtable) (vht: vp_hashtable) (deps: pkg_dep_expression): pkg_dep_expression =
begin
	List.map (fun disj ->
		List.flatten (List.map (fun dep ->
			if try (String.sub dep.target 0 6) = "rpmlib" with Invalid_argument _ -> false then []
			else if (!Options.version_type = Options.DebianVersions) && (dep.version_spec <> AnyVersion) then [dep] (* Debian provide resolution ONLY if dependency is not versioned *)
			else
			let possible_providers = Hashtbl.find_all vht dep.target in
			if !verbose then
				Printf.eprintf "[Deps] %s possibly provided by: %s\n" (Pretty_print.format_dependency dep) (String.concat "|" (List.map (fun (vs, d) -> "[" ^ (Pretty_print.format_dependency { target = d.target; version_spec = vs } ) ^ "]") possible_providers));
			if possible_providers = [] then
			begin
				Printf.eprintf "[Deps] WARNING: package %s-%s has a %s on %s, but %s does not exist.\n" name version dep_type (Pretty_print.format_dependency dep) dep.target; 
				Printf.eprintf "[Deps] Dropping %s, which might lead to inaccurate results (see manual).\n" dep_type 
			end;
			let providers = List.filter (fun (vs, _) -> can_provide vs dep.version_spec) possible_providers in
			if !verbose then
				Printf.eprintf "[Deps] %s finally provided by: %s\n" (Pretty_print.format_dependency dep) (String.concat "|" (List.map (fun (vs, d) -> "[" ^ (Pretty_print.format_dependency { target = d.target; version_spec = vs } ) ^ "]") providers));
			Tools.uniq (List.sort compare (List.map snd providers))
		) disj)
	) deps
end;;

let replace_virtual_conflicts (pht: pkg_hashtable) (vht: vp_hashtable) (pname: string) (cfls: pkg_dependency list): pkg_dependency list =
begin
	List.flatten (List.map (fun cfl ->
		if try (String.sub cfl.target 0 6) = "rpmlib" with Invalid_argument _ -> false then []
		else if (!Options.version_type = Options.DebianVersions) && (cfl.version_spec <> AnyVersion) then [cfl] (* Debian provide resolution ONLY if dependency is not versioned *)
		else
		let possible_providers = Hashtbl.find_all vht cfl.target in
		let providers = List.filter (fun (vs, d) -> can_provide vs cfl.version_spec && (if !Options.version_type = Options.DebianVersions then (d.target <> pname) && (cfl.target <> pname) else true)) possible_providers in
		let pkgs = Hashtbl.find_all pht cfl.target in
		let providers_plus =
		if !Options.version_type = Options.DebianVersions then
			(List.map snd providers) @ 
			(List.map
				(fun (v, m) -> { target = cfl.target; version_spec = (Equal v) })
				(List.filter (fun (v, _) ->
				can_provide (Equal v) cfl.version_spec) pkgs))
		else
			List.map snd providers in
				Tools.uniq (List.sort compare providers_plus)
	) cfls)
end;;

let resolve_dependencies (pkgs: pkg_metadata list): pkg_metadata list =
begin
	prerr_endline "[Deps] Resolving dependencies...";
	if !verbose then
		prerr_endline ("[Deps] Using " ^ (match !version_type with RPMVersions -> "RPM" | _ -> "Debian") ^ " versions...");
	let pht = create_pkgs_hashtable pkgs 
	and vht = create_vp_hashtable pkgs in
	List.map (fun pkg ->
		{ pkg with
			depends =
				replace_virtual_dependencies (pkg.name, pkg.version) "dependency" pht vht pkg.depends;
			pre_depends =
				replace_virtual_dependencies (pkg.name, pkg.version) "pre-dependency" pht vht pkg.pre_depends;
			conflicts = replace_virtual_conflicts pht vht pkg.name pkg.conflicts;
			recommends =
				replace_virtual_dependencies (pkg.name, pkg.version) "recommendation" pht vht pkg.recommends;
			suggests =
				replace_virtual_dependencies (pkg.name, pkg.version) "suggestion" pht vht pkg.suggests;
			enhances =
				replace_virtual_dependencies (pkg.name, pkg.version) "enhancement" pht vht pkg.enhances
		}
	) pkgs
end;;

let add_file_provides (pkgs: pkg_metadata list): pkg_metadata list =
let is_file_dependency (d: string): bool = d.[0] = '/' in
begin
 	if !verbose then prerr_endline "[Deps] Adding explicit file provides...";
	let filedep_ht = Hashtbl.create (List.length pkgs) in
	(* Scour packages for file dependencies *)
	List.iter (fun pkg ->
		List.iter (fun disj ->
			List.iter (fun dep ->
				if is_file_dependency dep.target then
					Hashtbl.add filedep_ht dep.target (pkg.name, pkg.version)
			) disj
		) pkg.depends;
		List.iter (fun disj ->
			List.iter (fun dep ->
				if is_file_dependency dep.target then
					Hashtbl.add filedep_ht dep.target (pkg.name, pkg.version)
			) disj
		) pkg.pre_depends;
		List.iter (fun disj ->
			List.iter (fun dep ->
				if is_file_dependency dep.target then
					Hashtbl.add filedep_ht dep.target (pkg.name, pkg.version)
			) disj
		) pkg.recommends;
		List.iter (fun disj ->
			List.iter (fun dep ->
				prerr_endline dep.target;
				if is_file_dependency dep.target then
					Hashtbl.add filedep_ht dep.target (pkg.name, pkg.version)
			) disj
		) pkg.suggests;
		List.iter (fun disj ->
			List.iter (fun dep ->
				if is_file_dependency dep.target then
					Hashtbl.add filedep_ht dep.target (pkg.name, pkg.version)
			) disj
		) pkg.enhances;
		List.iter (fun dep ->
			if is_file_dependency dep.target then
				Hashtbl.add filedep_ht dep.target (pkg.name, pkg.version)
		) pkg.replaces;
	) pkgs;
	(* Now add explicit file provides, but only for files that are needed *)
	List.map (fun pkg ->
		let extra_provides = List.flatten (List.map (fun (fn, _) ->
			let providers = Hashtbl.find_all filedep_ht fn in
			match providers with
			| [] -> []
			| _ -> [{ target = fn; version_spec = AnyVersion }]
		) pkg.files) in
		{ pkg with provides = pkg.provides @ extra_provides }
	) pkgs
end;;


let enlarge_cone (ht: pkg_hashtable) (vht: vp_hashtable) (units: (string * string * bool) list) : (string * string * bool) list =
let rec cone_uniq (x: (string * string * bool) list) (b: bool): (string * string * bool) list =
begin
	match x with
	| [] -> []
	| y::ys ->
		begin
			match ys with
			| [] -> let (ny, vy, dy) = y in [(ny, vy, b)]
			| z::zs -> let (ny, vy, dy) = y and (nz, vz, dz) = z in
				if ny = nz && vy = vz then
					cone_uniq ys (b || (dy || dz))
				else
					(ny, vy, dy)::(cone_uniq ys (b || dy))
		end
end in
begin
	let res = (List.flatten (List.flatten (List.map (fun (n, v, d) ->
	if not d then
	begin
		if (not (List.exists (fun (x, y) -> n = x && v = y) !Options.cone_stops)) then
		begin
		try
			let versions = Hashtbl.find_all ht n in
			let pkgs = List.map snd (if v = "" then versions else List.filter (fun (x, _) -> x = v) versions) in
			let deps = Tools.uniq (List.sort compare (List.flatten (List.map get_dependencies pkgs))) in
			List.map (fun dep ->
				let av_versions = Hashtbl.find_all ht dep.target in
				let providers = Hashtbl.find_all vht dep.target in
				List.flatten (List.map (fun (av_version, av_md) -> 
					if can_provide (Equal av_version) dep.version_spec then
						[(av_md.name, av_md.version, false)]	
					else
						[]
				) av_versions) @
				(List.flatten (List.map 
					(fun (vs, x) -> if can_provide dep.version_spec vs then
						[(x.target, version_of x.version_spec, false)]
					else
					[])
				providers))
			) deps
		with
			Not_found -> []
		end
		else
		begin
			prerr_endline ("[Deps] not enlarging for " ^ n ^ "/" ^ v ^ " because not allowed by user");
			[]
		end
	end
	else (* d is true *)
		[]
	) units))) in
	let new_units = List.map (fun (n,v,_) -> (n,v,true)) units in
	cone_uniq (List.sort (fun (n1,v1,d1) (n2,v2,d2) -> compare (n1,v1) (n2,v2)) (res @ new_units)) false
end;;

let extract_cone (pkgs: pkg_metadata list) (spec: string * string): pkg_metadata list =
let name, version = spec in
begin
	Printf.eprintf "[Deps] Extracting dependency cone for %s %s...\n" name (if version <> "" then " (version " ^ version ^")" else "");
	let pht = create_pkgs_hashtable pkgs
	and vht = create_vp_hashtable pkgs in
	let units = Tools.fixpoint [(name,version,false)] (enlarge_cone pht vht) in
	let result = List.flatten (List.map (fun (n, v, _) -> 
			let versions = Hashtbl.find_all pht n in
			List.map snd (if v = "" then versions else List.filter (fun (x, _) -> x = v) versions)
		) units) in
		prerr_endline ("[Deps] Cone size: " ^ (string_of_int (List.length result)));
		result
end;;
