(*
This file is part of Ceve.

Ceve is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Ceve is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Ceve; if not, write to the Free Software
Foundation, Inc., 52 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*)

open Metadata

let dose_of_architecture (arch: pkg_architecture): string =
begin
	match arch with
		AnyArchitecture -> "any"
	| AllArchitectures -> "all"
	| Source -> "source"
	| Specific s -> s
end;;

let dose_of_priority (priority: pkg_priority): string =
begin
	match priority with
		Required -> "required"
	| Important -> "important"
	| Standard -> "standard"
	| Optional -> "optional"
	| Extra -> "extra"
end;;

let dose_of_field (name: string) (contents: string): string * string * string list =
begin
	try
		let newline_index = String.index contents '\n' in
		let first_line = String.sub contents 0 newline_index in
		let next_lines = Str.split (Str.regexp_string "\n") (String.sub contents (newline_index + 1) ((String.length contents) - newline_index - 1)) in
			(name, first_line, next_lines)
	with Not_found -> (name, contents, [])
end;;

let dose_of_dependency (dep: pkg_dependency): string = 
begin
	dep.target ^ (match dep.version_spec with
		AnyVersion -> ""
	| Less v -> " (<< " ^ v ^ ")"
	| LessEqual v -> " (<= " ^ v ^ ")"
	| Equal v -> " (= " ^ v ^ ")"
	| Greater v -> " (>> " ^ v ^ ")"
	| GreaterEqual v -> " (>= " ^ v ^ ")"
	)
end;;

let dose_of_dependencies (deps: pkg_dependency list): string =
begin
	String.concat ", " (List.sort compare (Tools.uniq (List.map dose_of_dependency (List.filter (
		fun x -> not (Tools.starts_with "rpmlib" x.target))
	deps))))
end;;

let dose_of_dep_expression (deps: pkg_dep_expression): string =
begin
	String.concat ", " (List.sort compare (Tools.uniq (List.filter (fun z -> z <> "")
	(List.map (fun x ->
		String.concat " | " (List.sort compare (Tools.uniq (List.map dose_of_dependency (List.filter (fun y -> not (Tools.starts_with "rpmlib" y.target)) x))))
	) deps))))
end

let to_dose (pkg: pkg_metadata): (string * string * string list) list =
begin
	[
		("Architecture", dose_of_architecture pkg.architecture, []);
		dose_of_field "Bugs" pkg.bugs;
		("Conflicts", dose_of_dependencies pkg.conflicts, []);
		("Depends", dose_of_dep_expression pkg.depends, []);
		("Description", pkg.summary, Str.split (Str.regexp_string "\n") pkg.description);
		dose_of_field "X-Distribution" pkg.distribution;
		("Enhances", dose_of_dep_expression pkg.enhances, []);
		("Essential", (if pkg.essential then "yes" else "no"), []);
		dose_of_field "Filename" pkg.filename;
		("Installed-Size", Int32.to_string pkg.installed_size, []);
		dose_of_field "X-License" pkg.license;
		dose_of_field "Maintainer" pkg.maintainer;
		dose_of_field "MD5Sum" pkg.md5sum;
		dose_of_field "Package" pkg.name;
		dose_of_field "Origin" pkg.origin;
		dose_of_field "X-OS" pkg.os;
		("Pre-Depends", dose_of_dep_expression pkg.pre_depends, []);
		("Priority", dose_of_priority pkg.priority, []);
		("Provides", dose_of_dependencies (List.filter (fun x -> not (Tools.starts_with "rpmlib" x.target)) pkg.provides), []);
		("Recommends", dose_of_dep_expression pkg.recommends, []);
		("Replaces", dose_of_dependencies pkg.replaces, []);
		dose_of_field "Section" pkg.section;
		dose_of_field "SHA1" pkg.sha1sum;
		dose_of_field "SHA256" pkg.sha256sum;
		("Size", Int32.to_string pkg.size, []);
		("Source", (pkg.source_package ^ if pkg.source_version <> "" then "(" ^ pkg.source_version ^ ")" else ""), []);
		("Suggests", dose_of_dep_expression pkg.suggests, []);
		("Tags", String.concat ", " pkg.tags, []);
		("Task", String.concat ", " pkg.tasks, []);
		dose_of_field "url" pkg.url;
		dose_of_field "X-Vendor" pkg.vendor;
		dose_of_field "Version" pkg.version
	]
end;;

let output_dose (pkgs: pkg_metadata list): unit =
let module DBO = Dosebase.Out in
let db = DBO.open_out !Options.output_dir in
begin
	prerr_endline "[Dose] adding packages...";
	List.iter (fun pkg ->
		if !Options.verbose then
		begin
			prerr_endline ("[Dose] adding package " ^ pkg.name ^ ")")
		end;
	  DBO.add_package db ~archive:!Options.dose_archive ~day:(Lifetime.day_of_ymd !Options.dose_date) (to_dose pkg)
  ) pkgs;
	prerr_endline "[Dose] closing database...";
	DBO.close_out db
end;;
